%FFCC method implementation.
%input:
%        I         - input image in linear RGB space
%        ccm       - sensor ccm, which is used for image visualization
%        model     - learned filter
%        offset    - starting point of the log-chroma histogram
%        bin_width - bin size of the log-chroma histogram
%        bin_num   - number of the histogram bins
%output:
%        img_awb   - white balanecd image for visualization (ccm, gamma applied)
%        uv_est    - u, v values of the illuminant estimation

function [img_awb, uv_est] = FFCC(I, ccm, model, offset, bin_width, bin_num)
    
    %ignore pixels with zeros in any channel
    valid_mask = all(I > 0, 3);
    I_dev = MaskedLocalAbsoluteDeviation(I, valid_mask);
    
    %compute log-chroma histogram
    %hist1 - histogram of the log-chroma of the original image
    %hist2 - histogram of the log-chroma of the local absolute deviation
    
    [hist1, uv_mean] = LogChromaHist(I, valid_mask, offset, bin_width, bin_num);
    hist2 = LogChromaHist(I_dev, valid_mask, offset, bin_width, bin_num);
    hist = cat(3, hist1, hist2);
    
    %vis = VisualizeHistogram(hist1+hist2, params);
    %figure, imshow(vis);
    
    %convolution histogram with learned filter F
    hist_fft = fft2(hist);
    filtered_hist = real(ifft2(sum(model.F_fft .* hist_fft, 3))) + model.B;

    %softmax to convert filtered hist to PDF
    P = SoftmaxForward(filtered_hist);

    %fit bivariate Von Mises to compute the centroid of the filtered histogram
    uv_est_idx = BivariateVoMises(P);
    uv_est = (uv_est_idx - 1) * bin_width + offset;
    
    tmp = bin_width * bin_num;
    %uv_est = uv_est - tmp * floor(1/tmp * (uv_est - uv_mean) + 1/2);

    % Convert the white point into a RGB gain.
    gains = [exp(uv_est(1)); 1; exp(uv_est(2))];
    gains = gains ./ min(gains(:));
    
    img_awb = VisalizeImg(I, ccm, gains);
    
end